/*
 Custom Logo Appender
 Version: 1.1
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 29/11/2019
 Last Modified: 22/06/2023
 License: CC-BY-NC
  
 Drag and drop both of ROM files. This is for the 1MB Ultra Low Power Cart.
 
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <libgen.h>

#ifdef _WIN32
#include <Windows.h>
#include <direct.h>
#define GetCurrentDir _getcwd
#else
#include <unistd.h>
#define GetCurrentDir getcwd
#endif

uint8_t nintendoLogo[] = {0xCE, 0xED, 0x66, 0x66, 0xCC, 0x0D, 0x00, 0x0B, 0x03, 0x73, 0x00, 0x83, 0x00, 0x0C, 0x00, 0x0D,
									0x00, 0x08, 0x11, 0x1F, 0x88, 0x89, 0x00, 0x0E, 0xDC, 0xCC, 0x6E, 0xE6, 0xDD, 0xDD, 0xD9, 0x99,
									0xBB, 0xBB, 0x67, 0x63, 0x6E, 0x0E, 0xEC, 0xCC, 0xDD, 0xDC, 0x99, 0x9F, 0xBB, 0xB9, 0x33, 0x3E};

// Read one letter from stdin
char read_one_letter (void) {
	char c = getchar();
	while (getchar() != '\n' && getchar() != EOF);
	return c;
}

int main(int argc, char **argv) {
	
	printf("insideGadgets - Custom Logo Appender v1.1\n");
	printf("#########################################\n");
	printf("The resulting file can only be used on insideGadgets Gameboy Flash Carts that support it.\n\n");
	
	if (argc >= 3) {
		char outputFile[255];
		
		// Read the logo from both files
		FILE *romReadFile1 = fopen(argv[1], "rb");
		fseek (romReadFile1, 0x104, SEEK_SET);
		uint8_t romFileLogo1[48];
		fread(romFileLogo1, 1, 48, romReadFile1);
		fseek (romReadFile1, 0, SEEK_END);
		long fileSize1 = ftell(romReadFile1);
		fclose(romReadFile1);
		
		FILE *romReadFile2 = fopen(argv[2], "rb");
		fseek (romReadFile2, 0x104, SEEK_SET);
		uint8_t romFileLogo2[48];
		fread(romFileLogo2, 1, 48, romReadFile2);
		fseek (romReadFile2, 0, SEEK_END);
		long fileSize2 = ftell(romReadFile1);
		fclose(romReadFile2);
		
		// Check file for logo
		uint8_t logoCheck = 1;
		for (uint16_t logoAddress = 0; logoAddress < 48; logoAddress++) {
			if (nintendoLogo[logoAddress] != romFileLogo1[logoAddress]) {
				logoCheck = 0;
				break;
			}
		}
		if (logoCheck == 0) {
			logoCheck = 2;
			for (uint16_t logoAddress = 0; logoAddress < 48; logoAddress++) {
				if (nintendoLogo[logoAddress] != romFileLogo2[logoAddress]) {
					logoCheck = 0;
					break;
				}
			}
		}
		
		// First file has valid logo
		if (logoCheck == 1) {
			// Strip file extension
			char romMapFileName[250];
			strncpy(romMapFileName, argv[1], strlen(argv[1])-3);
			romMapFileName[strlen(argv[1])-3] = 0;
			
			// New file
			strncpy(outputFile, romMapFileName, 220);
			strncat(outputFile, "_custom_logo.gb", 16);
			romReadFile1 = fopen(argv[1], "rb");
			FILE *romWriteFile = fopen(outputFile, "wb");
			uint8_t buffer[1] = {0xFF};
			
			// Copy 0x0000-0x3FFF from the existing file to a buffer
			uint8_t bank0Buffer[0x4000];
			fread(bank0Buffer, 1, 0x4000, romReadFile1);
			
			uint8_t bankSwitchFix[35] = {0xF5, 0x3E, 0x3E, 0x21, 0x80, 0xFF, 0x22, 0x3E, 0x01, 0x22, 0x3E, 0xEA, 0x22, 0x3E, 0x00, 0x22, 0x3E, 0x20, 0x22, 0x3E, 0xF1, 0x22, 0x3E, 0xC3, 0x22, 0x3E, 0x00, 0x22, 0x3E, 0x01, 0x22, 0xC3, 0x80, 0xFF};
			for (uint8_t x = 0; x < 34; x++) {
				bank0Buffer[0x150 + x] = bankSwitchFix[x];
			}
			bank0Buffer[0x100] = 0x00;
			bank0Buffer[0x101] = 0xC3;
			bank0Buffer[0x102] = 0x50;
			bank0Buffer[0x103] = 0x01;
			
			// Copy contents of original file to the new file
			if (fileSize1 > 1032192) {
				fileSize1 = 1032192; // Limit to 1008KB
			}
			fseek (romReadFile1, 0, SEEK_SET);
			for (uint32_t x = 0; x < fileSize1; x++) {
				fread(buffer, 1, 1, romReadFile1);
				fwrite(buffer, 1, 1, romWriteFile);
			}
			
			// Write 0xFFs until the size is 1,032,192 bytes
			for (uint32_t x = fileSize1; x < 1032192; x++) {
				buffer[0] = 0xFF;
				fwrite(buffer, 1, 1, romWriteFile);
			}
			
			// Overwrite the logo area with the other file's logo
			for (uint16_t logoAddress = 0x104; logoAddress <= 0x133; logoAddress++) {
				bank0Buffer[logoAddress] = romFileLogo2[(logoAddress-0x104)];
			}
			
			// Jump to 1008KB and write the first bank with the custom logo
			fseek (romReadFile1, 1032192, SEEK_SET);
			fwrite(bank0Buffer, 1, 0x4000, romWriteFile);
			
			fclose(romReadFile1);
			fclose(romWriteFile);
			
			printf("Completed.\n");
		}
		else if (logoCheck == 2) {
			// Strip file extension
			char romMapFileName[250];
			strncpy(romMapFileName, argv[2], strlen(argv[2])-3);
			romMapFileName[strlen(argv[2])-3] = 0;
			
			// New file
			strncpy(outputFile, romMapFileName, 220);
			strncat(outputFile, "_custom_logo.gb", 16);
			romReadFile2 = fopen(argv[2], "rb");
			FILE *romWriteFile = fopen(outputFile, "wb");
			uint8_t buffer[1] = {0xFF};
			
			// Copy 0x0000-0x3FFF from the existing file to a buffer
			uint8_t bank0Buffer[0x4000];
			fread(bank0Buffer, 1, 0x4000, romReadFile2);
			
			uint8_t bankSwitchFix[35] = {0xF5, 0x3E, 0x3E, 0x21, 0x80, 0xFF, 0x22, 0x3E, 0x01, 0x22, 0x3E, 0xEA, 0x22, 0x3E, 0x00, 0x22, 0x3E, 0x20, 0x22, 0x3E, 0xF1, 0x22, 0x3E, 0xC3, 0x22, 0x3E, 0x00, 0x22, 0x3E, 0x01, 0x22, 0xC3, 0x80, 0xFF};
			for (uint8_t x = 0; x < 34; x++) {
				bank0Buffer[0x150 + x] = bankSwitchFix[x];
			}
			bank0Buffer[0x100] = 0x00;
			bank0Buffer[0x101] = 0xC3;
			bank0Buffer[0x102] = 0x50;
			bank0Buffer[0x103] = 0x01;
			
			// Copy contents of original file to the new file
			if (fileSize2 > 1032192) {
				fileSize2 = 1032192; // Limit to 1008KB
			}
			fseek (romReadFile2, 0, SEEK_SET);
			for (uint32_t x = 0; x < fileSize2; x++) {
				fread(buffer, 1, 1, romReadFile2);
				fwrite(buffer, 1, 1, romWriteFile);
			}
			
			// Write 0xFFs until the size is 1,032,192 bytes
			for (uint32_t x = fileSize2; x < 1032192; x++) {
				buffer[0] = 0xFF;
				fwrite(buffer, 1, 1, romWriteFile);
			}
			
			// Overwrite the logo area with the other file's logo
			for (uint16_t logoAddress = 0x104; logoAddress <= 0x133; logoAddress++) {
				bank0Buffer[logoAddress] = romFileLogo1[(logoAddress-0x104)];
			}
			
			// Jump to 1008KB and write the first bank with the custom logo
			fseek (romReadFile2, 1032192, SEEK_SET);
			fwrite(bank0Buffer, 1, 0x4000, romWriteFile);
			
			fclose(romReadFile2);
			fclose(romWriteFile);
			
			printf("Completed.\n");
		}
		else {
			printf("Neither of the rom files have a valid Nintendo Logo. Won't continue.\n");
			read_one_letter();
			return 1;
		}
	}
	else {
		printf("Please drag and drop 2 files or pass them as agruments.");
	}
	
	read_one_letter();
	
	return 0;
}